
'''

Solve a Poisson transmission problem with mixed FEM

- \Delta p_top    = f_top      in \Omega_top    := (0,1)x(0.5, 1)
- \Delta p_bottom = f_bottom   in \Omega_bottom := (0,1)x(0, 0.5)

                        p_top - p_bottom = jump1                on interface
         \nabla p_top \cdot n - \nabla p_bottom \cdot n = jump2 on interface 

The exact solution is: 

    p_top    = sin(pi*x)*sin(pi*y) 
and 
    p_bottom = 8*x*(1-x)*y*(1-y)

'''

from ngsolve import *
from netgen.geom2d import SplineGeometry

from numpy import pi


#Exact solution + source terms + jumps on interface

p_top = sin(pi*x)*sin(pi*y)
p_bottom = 8*x*(1-x)*y*(1-y)

ftop = 2*pi**2*p_top
fbottom = 16*y*(1-y) + 16*x*(1-x)

gradp_top = CoefficientFunction( (pi*cos(pi*x)*sin(pi*y), pi*sin(pi*x)*cos(pi*y)) )
gradp_bottom = CoefficientFunction( ( 8*(1-2*x)*y*(1-y),         8*x*(1-x)*(1-2*y)) )

jump1 = p_top - p_bottom
jump2 = gradp_top - gradp_bottom

# Geometry and mesh

def topBottom():
    geometry = SplineGeometry()
    pnts     = [ (0, 0), (1, 0),  (1, 0.5), (1, 1), (0, 1), (0, 0.5)]
    pnums    = [geometry.AppendPoint(*p) for p in pnts]
    # start-point, end-point, boundary-condition, domain on left side, domain on right side:
    lines = [ (pnums[0], pnums[1], "gammaBottom",  2, 0),
              (pnums[1], pnums[2], "gammaBottom",  2, 0), 
              (pnums[2], pnums[3], "gammaTop",     1, 0),
              (pnums[3], pnums[4], "gammaTop",     1, 0),
              (pnums[4], pnums[5], "gammaTop",     1, 0),
              (pnums[5], pnums[0], "gammaBottom",  2, 0),
              (pnums[5], pnums[2], "interface",    1, 2)]
    for p1,p2,bc,left,right in lines:
        geometry.Append( ["line", p1, p2], bc=bc, leftdomain=left, rightdomain=right)

    geometry.SetMaterial(1, "top")
    geometry.SetMaterial(2, "bottom")
        
    return geometry


mesh = Mesh(topBottom().GenerateMesh (maxh=0.1))

n = specialcf.normal(mesh.dim)

Draw(mesh)

# Mixed finite element spaces

order = 4

V1 = HDiv(mesh, order = order,   definedon="top")
M1 = L2(mesh,   order = order-1, definedon="top")
V2 = HDiv(mesh, order = order,   definedon="bottom")
M2 = L2(mesh,   order = order-1, definedon="bottom")
L  = L2(mesh,   order = order,   definedon="", definedonbound="interface")

X = FESpace([V1, V2, M1, M2, L])
u1, u2, p1, p2, lmbda = X.TrialFunction()
v1, v2, q1, q2, mu    = X.TestFunction()

# Bilinear and linear forms

a = BilinearForm(X, symmetric=True)
a += u1*v1*dx('top') + u2*v2*dx('bottom')
a += div(u1)*q1*dx('top')    + div(v1)*p1*dx('top')
a += div(u2)*q2*dx('bottom') + div(v2)*p2*dx('bottom')

a += (u2.Trace()*n - u1.Trace()*n)*mu   *ds('interface')
a += (v2.Trace()*n - v1.Trace()*n)*lmbda*ds('interface')


f = LinearForm(X)
f += - ftop * q1 * dx('top') - fbottom * q2 * dx('bottom')
f += jump1*(v1.Trace()*n)*ds('interface')
f +=  -(jump2*n)*mu*ds('interface')

#Solving the problem
a.Assemble()
f.Assemble()

u = GridFunction(X)
u.vec.data = a.mat.Inverse(X.FreeDofs()) * f.vec

gfp1 = u.components[2]
gfp2 = u.components[3]

Vpost = L2(mesh, order = order-1)
glob = GridFunction(Vpost, name="aprox")
glob.Set(gfp1+gfp2) # The global solution

# project the exact solution to  L2 space

up = Vpost.TrialFunction()
vp = Vpost.TestFunction()
ap = BilinearForm(Vpost)
ap += SymbolicBFI(up*vp)

ap.Assemble()

fpe = LinearForm(Vpost)
fpe += p_top*vp*dx('top')
fpe += p_bottom*vp*dx('bottom')
fpe.Assemble()

upe = GridFunction(Vpost, name="exact")
upe.vec.data = ap.mat.Inverse()*fpe.vec

# L2 error
err = sqrt(Integrate((upe-glob)*(upe-glob),mesh))
print(err)

Draw(upe)
Draw(glob)